/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

@namespace url("http://www.w3.org/1999/xhtml");
@namespace xul url("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul");

/* Style the scrollbars */
xul|window xul|scrollbar {
  display: none;
}

html xul|scrollbar {
  display: block;
}

xul|scrollbar[orient="vertical"] {
  -moz-appearance: none !important;
  position: relative;
  margin-left: -8px;
  min-width: 8px;
  max-width: 8px;
  background-color: transparent !important;
  background-image: none !important;
  border: 0px solid transparent !important;
}

xul|scrollbar[orient="vertical"]:-moz-locale-dir(rtl) {
  margin-left: 2px;
  margin-right: -10px;
}

xul|scrollbar[orient="vertical"] xul|thumb {
  max-width: 6px !important;
  min-width: 6px !important;
}

xul|scrollbar[orient="horizontal"] {
  -moz-appearance: none !important;
  position: relative;
  min-height: 8px;
  max-height: 8px;
  margin-top: -8px;
  background-color: transparent !important;
  background-image: none !important;
  border: 0px solid transparent !important;
}

xul|scrollbar[orient="horizontal"] xul|thumb {
  max-height: 6px !important;
  min-height: 6px !important;
}

xul|scrollbox {
  overflow-y: scroll;
  overflow-x: scroll;
}

xul|scrollbarbutton {
  min-height: 8px !important;
  min-width: 8px !important;
  -moz-appearance: none !important;
  visibility: hidden;
}

xul|scrollbarbutton[sbattr="scrollbar-up-top"],
xul|scrollbarbutton[sbattr="scrollbar-bottom-top"] {
  display: none;
}

xul|thumb {
  background-color: rgba(0, 0, 0, 0.4) !important;
  -moz-border-top-colors: none !important;
  -moz-border-bottom-colors: none !important;
  -moz-border-right-colors: none !important;
  -moz-border-left-colors: none !important;
  border: 1px solid rgba(255, 255, 255, 0.4) !important;
  border-radius: 3px;
}

/* -moz-touch-enabled? media elements */
:-moz-any(video, audio) > xul|videocontrols {
  -moz-binding: url("chrome://global/content/bindings/videocontrols.xml#touchControls");
}
